#include "mesh_augmentor.h"

#include "Eigen/Core"

#include "geometrycentral/surface/manifold_surface_mesh.h"
#include "geometrycentral/surface/surface_mesh.h"
#include "geometrycentral/surface/surface_mesh_factories.h"

#include <earcut.hpp>

#include <algorithm>
#include <array>
#include <chrono>
#include <random>
#include <vector>

typedef Eigen::Matrix<double, 3, 1> Float3;
typedef Eigen::Matrix<int32_t, 3, 1> Int3;

using namespace geometrycentral;
using namespace geometrycentral::surface;


MeshAugmentor::MeshAugmentor(ManifoldSurfaceMesh& mesh_, VertexPositionGeometry& geom_, MeshAugmentorOptions options_)
    : mesh(mesh_), geom(geom_), options(options_),
      rng(std::chrono::high_resolution_clock::now().time_since_epoch().count()), corner_pos_2d(mesh) {}


void MeshAugmentor::run() {
  build_augumentation_list();
  process_augumentation_list();
  mesh.compress();
}

void MeshAugmentor::build_augumentation_list() {
  augment_pipeline.clear();

  { // face subdiv
    size_t N = static_cast<size_t>(std::ceil(mesh.nFaces() * options.face_subdiv_rate));
    for (size_t i = 0; i < N; i++) augment_pipeline.push_back(AugmentOp::FaceSubdiv);
  }

  { // edge split
    size_t N = static_cast<size_t>(std::ceil(mesh.nEdges() * options.edge_split_rate));
    for (size_t i = 0; i < N; i++) augment_pipeline.push_back(AugmentOp::EdgeSplit);
  }

  { // face split
    size_t N = static_cast<size_t>(std::ceil(mesh.nFaces() * options.face_split_rate));
    for (size_t i = 0; i < N; i++) augment_pipeline.push_back(AugmentOp::FaceSplit);
  }

  { // planar edge flip
    size_t N = static_cast<size_t>(std::ceil(mesh.nEdges() * options.planar_edge_flip_rate));
    for (size_t i = 0; i < N; i++) augment_pipeline.push_back(AugmentOp::PlanarEdgeFlip);
  }


  // { // planar vertex perturb
  //   size_t N = static_cast<size_t>(std::ceil(mesh.nVertices() * options.planar_vertex_perturb_rate));
  //   for (size_t i = 0; i < N; i++) augment_pipeline.push_back(AugmentOp::PlanarVertexPerturb);
  // }

  // shuffle the list
  std::shuffle(augment_pipeline.begin(), augment_pipeline.end(), rng);
}


void MeshAugmentor::process_augumentation_list() {
  for (const AugmentOp& op : augment_pipeline) {

    // dispatch to the appropriate op
    switch (op) {
    case AugmentOp::FaceSubdiv:
      do_face_subdiv();
      break;
    case AugmentOp::EdgeSplit:
      do_edge_split();
      break;
    case AugmentOp::FaceSplit:
      do_face_split();
      break;

    case AugmentOp::PlanarEdgeFlip:
      do_planar_edge_flip();
      break;

      // case AugmentOp::PlanarVertexPerturb:
      //   do_planar_vertex_perturb();
      //   break;
    }
  }
}


void MeshAugmentor::do_face_subdiv() {

  // Pick a random face
  std::uniform_int_distribution<size_t> distrib_int(0, mesh.nFaces() - 1);
  size_t target_face_ind = distrib_int(rng);
  Face target_face = mesh.face(target_face_ind);

  if(!face_is_convex(target_face)) return;

  // Randomly barycentric coordinates within the face
  // (note that we handle the general polygonal case here)
  std::vector<double> bCoords;
  double bCoordsSum = 0.;
  std::uniform_real_distribution<double> distrib_real(0.1, 0.9);
  for (Vertex v : target_face.adjacentVertices()) {
    double c = distrib_real(rng);
    bCoords.push_back(c);
    bCoordsSum += c;
  }
  for (double& c : bCoords) c /= bCoordsSum; // normalize to sum to 1


  // Generate a new position for the vertex
  Vector3 new_pos{0., 0., 0.};
  size_t i = 0;
  for (Vertex v : target_face.adjacentVertices()) {
    double c = bCoords[i];
    Vector3 v_pos = geom.vertexPositions[v];
    new_pos += c * v_pos;
    i++;
  }

  Vector3 normal_before = geom.faceNormal(target_face);

  // Insert the new vertex
  Vertex new_v = mesh.insertVertex(target_face);
  geom.vertexPositions[new_v] = new_pos;

  // Test if the new geometry we created significantly changes the normal of any of the faces. If so this was a very
  // nonplanar face. Or even worse, it could be a nonconvex face and we created a flip. If so, undo the operatoin.
  if (!target_face.isTriangle()) { // triangles are always planar and convex == safe
    bool must_undo = false;

    for (Face neigh_f : new_v.adjacentFaces()) {
      Vector3 new_normal = geom.faceNormal(neigh_f);
      if (!(dot(normal_before, new_normal) >= (1. - options.planar_threshold - options.rel_eps))) {
        must_undo = true;
      }
    }

    if (must_undo) {
      // this should only happen rarely
      mesh.removeVertex(new_v);
      mesh.compress(); // expensive!
    }
  }
}

void MeshAugmentor::do_edge_split() {

  // Pick a random edge
  std::uniform_int_distribution<size_t> distrib_int(0, mesh.nEdges() - 1);
  size_t target_edge_ind = distrib_int(rng);
  Edge target_edge = mesh.edge(target_edge_ind);

  // Randomly barycentric coordinates within the face
  std::uniform_real_distribution<double> distrib_real(0.1, 0.9);
  double c = distrib_real(rng);

  // Generate a new position for the vertex
  // clang-format off
  Vector3 new_pos =
             c * geom.vertexPositions[target_edge.firstVertex()] + 
      (1. - c) * geom.vertexPositions[target_edge.secondVertex()];

  // Check if the faces on either side are triangles
  bool neighborhood_triangular =
      // first face is triangular
      target_edge.halfedge().face().isTriangle()
      && // AND
      // second face boundary, or is triangular
      (
       !target_edge.halfedge().twin().isInterior() || 
       target_edge.halfedge().twin().face().isTriangle()
      )
    ;

  // Insert the new vertex
  // If the neighborhood is triangular, use the routine 
  // which triangulates after the insertion
  Vertex new_v = neighborhood_triangular ? 
        mesh.splitEdgeTriangular(target_edge).vertex() : 
        mesh.insertVertexAlongEdge(target_edge).vertex();
  // clang-format on

  geom.vertexPositions[new_v] = new_pos;
}

void MeshAugmentor::do_face_split() {

  // Pick a random face
  std::uniform_int_distribution<size_t> distrib_int(0, mesh.nFaces() - 1);
  size_t target_face_ind = distrib_int(rng);
  Face target_face = mesh.face(target_face_ind);

  // If it's triangular, we can't do anything
  if (target_face.isTriangle()) {
    return;
  }
  int32_t D = target_face.degree();

  // Triangulate the face
  // We will pick an internal edge from this triangulation as the new edge. This leverages triangulation logic and saves
  // us from needing to check valid splits for nonconvex polygons, which is nontrivial.
  // TODO this gives us a kinda weird distribution of of splits, are we okay with that?

  update_corner_pos_2d(target_face);

  // copy over to earcut data structures
  using EarcutPoint = std::array<double, 2>;
  std::vector<std::vector<EarcutPoint>> polygon;
  // first list is the main polygon, subsequent lists are holes, if we had them
  polygon.emplace_back();
  for (Corner c : target_face.adjacentCorners()) {
    polygon.back().push_back({corner_pos_2d[c].x, corner_pos_2d[c].y});
  }

  // run the earcut clipping algorithm
  std::vector<uint32_t> tri_inds_flat = mapbox::earcut<uint32_t>(polygon);

  if (tri_inds_flat.empty()) return; // clipping failed

  // copy to list of triples
  std::vector<std::array<uint32_t, 3>> tri_inds;
  tri_inds.reserve(tri_inds_flat.size() / 3);
  for (int32_t i = 0; i < tri_inds_flat.size(); i += 3) {
    tri_inds.push_back({tri_inds_flat[i + 0], tri_inds_flat[i + 1], tri_inds_flat[i + 2]});
  }

  // shuffle the list
  std::shuffle(tri_inds.begin(), tri_inds.end(), rng);

  // take the first internal edge
  int32_t eA = -1;
  int32_t eB = -1;
  for (int32_t i = 0; i < 3; i++) {
    if (((tri_inds[0][i] + 1 + D) % D) != tri_inds[0][(i + 1) % 3]) {
      // this is an internal edge
      eA = tri_inds[0][i];
      eB = tri_inds[0][(i + 1) % 3];
    }
  }
  if (eA == -1 || eB == -1) return; // failure of earclip

  // find the corners corresponding to those endpoints
  Corner cA, cB;
  int32_t iC = 0;
  for (Corner c : target_face.adjacentCorners()) {
    if (iC == eA) cA = c;
    if (iC == eB) cB = c;
    iC++;
  }

  // Connect them and update the mesh
  Halfedge newHe = mesh.connectVertices(cA.halfedge(), cB.halfedge());
  if (newHe == Halfedge()) {
    throw std::runtime_error("could not connect vertices in do_face_split()");
  }
}

void MeshAugmentor::do_planar_edge_flip() {

  // Pick a random edge
  std::uniform_int_distribution<size_t> distrib_int(0, mesh.nEdges() - 1);
  size_t target_edge_ind = distrib_int(rng);
  Edge target_edge = mesh.edge(target_edge_ind);

  // Quick-out if we can't flip
  if (target_edge.isBoundary()) {
    return;
  }
  Face face_A = target_edge.halfedge().face();
  Face face_B = target_edge.halfedge().twin().face();
  if (!face_A.isTriangle() || !face_B.isTriangle()) {
    return;
  }
  Vector3 normal_A = geom.faceNormal(face_A);
  Vector3 normal_B = geom.faceNormal(face_B);
  Vector3 neigh_normal = unit(normal_A + normal_B);

  { // check planarity condition
    bool inplane_A = dot(neigh_normal, normal_A) >= (1. - options.planar_threshold - options.rel_eps);
    bool inplane_B = dot(neigh_normal, normal_B) >= (1. - options.planar_threshold - options.rel_eps);
    if (!inplane_A || !inplane_B) {
      return;
    }
  }

  // do the flip
  bool flipped = mesh.flip(target_edge);
  if (!flipped) return;

  // recompute those normals, they may have changed
  // (the face entries are re-used)
  normal_A = geom.faceNormal(face_A);
  normal_B = geom.faceNormal(face_B);

  { // check the planarity condition again, in case we created an inverted face or something else went wrong. If so,
    // undo the flip.
    bool inplane_A = dot(neigh_normal, normal_A) >= (1. - options.planar_threshold - options.rel_eps);
    bool inplane_B = dot(neigh_normal, normal_B) >= (1. - options.planar_threshold - options.rel_eps);
    if (!inplane_A || !inplane_B) {
      // undo the flip
      mesh.flip(target_edge);
    }
  }
}


bool MeshAugmentor::face_is_convex(Face face) {
  if (face.isTriangle()) return true;

  update_corner_pos_2d(face);

  for (Halfedge he : face.adjacentHalfedges()) {

    Corner c_prev = he.corner();
    Corner c_middle = he.next().corner();
    Corner c_next = he.next().next().corner();

    Vector2 p_prev = corner_pos_2d[c_prev];
    Vector2 p_middle = corner_pos_2d[c_middle];
    Vector2 p_next = corner_pos_2d[c_next];

    double angle = orientedAngle(p_middle - p_prev, p_next - p_middle);

    if (angle < 0) return false;
  }

  return true;
}


void MeshAugmentor::update_corner_pos_2d(Face face) {

  // Project to 2D
  Vector3 normal = geom.faceNormal(face);
  std::array<Vector3, 2> basis = normal.buildTangentBasis();
  Vector3 basisX = basis[0];
  Vector3 basisY = basis[1];

  for (Corner c : face.adjacentCorners()) {
    Vector3 pos = geom.vertexPositions[c.vertex()];
    Vector2 pos2d{dot(pos, basisX), dot(pos, basisY)};
    corner_pos_2d[c] = pos2d;
  }
}
