#pragma once

#include <numeric>

#include "Eigen/Core"

#include "geometrycentral/surface/manifold_surface_mesh.h"
#include "geometrycentral/surface/surface_mesh.h"
#include "geometrycentral/surface/surface_mesh_factories.h"

typedef Eigen::Matrix<double, 3, 1> Float3;
typedef Eigen::Matrix<int32_t, 3, 1> Int3;

namespace gc = geometrycentral;

struct MeshAugmentorOptions {

  // These rates are all expressed relative to the number of elemnets in the mesh. So a face_subdiv_rate of 0.1 means
  // "do 0.1*num_faces splits".

  double face_subdiv_rate = 0.05;
  double edge_split_rate = 0.05;
  double face_split_rate = 0.1;
  double planar_edge_flip_rate = 0.1;
  // double planar_vertex_perturb_rate = 0.05;


  // Smaller values mean a more strict planarity test.
  // Precisely, dot product between normal vectors must be at least
  // 1.-planar_threshold in order for planar operations to be allowed.
  double planar_threshold = 0.05;

  double rel_eps = 1e-6; // a numerical epsilon used in some tests
};

enum class AugmentOp { FaceSubdiv = 0, EdgeSplit, FaceSplit, PlanarEdgeFlip, /* PlanarVertexPerturb */ };

class MeshAugmentor {
public:
  MeshAugmentor(gc::surface::ManifoldSurfaceMesh& mesh, gc::surface::VertexPositionGeometry& geom,
                MeshAugmentorOptions options);

  // input arguments
  gc::surface::ManifoldSurfaceMesh& mesh;
  gc::surface::VertexPositionGeometry& geom;
  MeshAugmentorOptions options;

  // internal state
  std::mt19937 rng;

  // high level
  void run();

  // augmentation pipline
  std::vector<AugmentOp> augment_pipeline;
  void build_augumentation_list();
  void process_augumentation_list();

  // individual augmentation ops
  void do_face_subdiv();
  void do_edge_split();
  void do_face_split();
  void do_planar_edge_flip();
  // void do_planar_vertex_perturb();

  // shared helper
  bool face_is_convex(gc::surface::Face face);

  // Populate a temporary buffer of a face-corner's position in a normal coordinate system for the face
  // These are stored in corner_pos_2d. The buffer is *not* always kept up to date, it must temporarily updated via
  // update_corner_pos_2d() then immediately consumed. NOTE NOT THREADSAFE
  gc::surface::CornerData<gc::Vector2> corner_pos_2d;
  void update_corner_pos_2d(gc::surface::Face face);
};
