#include <numeric>

#include "mesh_augmentor.h"

#include "Eigen/Core"

#include <nanobind/eigen/dense.h>
#include <nanobind/nanobind.h>
#include <nanobind/stl/string.h>
#include <nanobind/stl/tuple.h>
#include <nanobind/stl/vector.h>

#include "igl/fast_find_self_intersections.h"

#include "geometrycentral/surface/manifold_surface_mesh.h"
#include "geometrycentral/surface/meshio.h"
#include "geometrycentral/surface/simple_polygon_mesh.h"
#include "geometrycentral/surface/surface_mesh.h"
#include "geometrycentral/surface/surface_mesh_factories.h"
#include "geometrycentral/utilities/eigen_interop_helpers.h"
#include "geometrycentral/utilities/elementary_geometry.h"
#include "geometrycentral/surface/heat_method_distance.h"

#include <earcut.hpp>


namespace nb = nanobind;
using namespace nb::literals;

typedef Eigen::Matrix<double, 3, 1> Float3;
typedef Eigen::Matrix<int32_t, 3, 1> Int3;

namespace gc = geometrycentral;
using gc::Vector2;
using gc::Vector3;

// ======================================================================
// ==== helper functions
// ======================================================================

// warning: with the current algorithm this sometimes outputs fewer than expected triangles
void triangulate_polygon_ear_clip_mapbox(std::vector<Vector3> vertex_positions, std::vector<Int3>& tri_inds_out) {

  tri_inds_out.clear();
  int32_t nC = vertex_positions.size();

  // quick-out for triangles
  if (nC == 3) {
    tri_inds_out.push_back({0, 1, 2});
    return;
  }

  // estimate a normal direction
  // average of corner normals (note that concave edges contribute an opposite
  // normal, but these will be outweight by the correct-facing ones)
  Vector3 plane_normal = Vector3::zero();
  {
    for (int32_t iC = 0; iC < nC; iC++) {
      int32_t iC_prev = (iC + nC - 1) % nC;
      int32_t iC_next = (iC + nC + 1) % nC;
      Vector3 pos_prev = vertex_positions[iC_prev];
      Vector3 pos = vertex_positions[iC];
      Vector3 pos_next = vertex_positions[iC_next];
      Vector3 vecA = (pos - pos_prev);
      Vector3 vecB = (pos_next - pos);
      Vector3 thisN = gc::cross(vecA, vecB);
      plane_normal += thisN;
    }
    plane_normal.normalize();
  }

  // Project the positions to 2d
  std::vector<Vector2> vertex_positions2d(vertex_positions.size());
  {

    std::array<Vector3, 2> basis = plane_normal.buildTangentBasis();
    Vector3 basisX = basis[0];
    Vector3 basisY = basis[1];

    int32_t nC = vertex_positions.size();
    for (int32_t iC = 0; iC < nC; iC++) {
      Vector3 pos = vertex_positions[iC];
      vertex_positions2d[iC] = {dot(pos, basisX), dot(pos, basisY)};
    }
  }

  // copy over to earcut data structures
  using EarcutPoint = std::array<double, 2>;
  std::vector<std::vector<EarcutPoint>> polygon;
  // first list is the main polygon, subsequent lists are holes, if we had them
  polygon.emplace_back();
  for (int32_t iC = 0; iC < nC; iC++) {
    polygon.back().push_back({vertex_positions2d[iC].x, vertex_positions2d[iC].y});
  }

  // run the earcut clipping algorithm
  std::vector<uint32_t> tri_inds_flat = mapbox::earcut<uint32_t>(polygon);

  // copy back in to our data structure
  for (int32_t i = 0; i < tri_inds_flat.size(); i += 3) {
    tri_inds_out.push_back({(int)tri_inds_flat[i + 0], (int)tri_inds_flat[i + 1], (int)tri_inds_flat[i + 2]});
  }

  // std::cout << "poly size = " << polygon[0].size() << " flat size = " << tri_inds_flat.size()
  //           << " size = " << tri_inds_out.size() << " nC = " << nC << std::endl;
  // if (tri_inds_out.size() != nC - 2) throw std::runtime_error("ear clip not enough tris");
}


void triangulate_polygon_ear_clip_greedy(std::vector<Float3>& vertex_positions_mut, std::vector<Int3>& tri_inds_out) {

  // NOTE: code adapted from autoUV repo
  // this uses an inline greedy algorithm

  tri_inds_out.clear();

  // estimate a normal direction
  // average of corner normals (note that concave edges contribute an opposite
  // normal, but these will be outweight by the correct-facing ones)
  Float3 plane_normal = Float3::Zero();
  {
    int32_t nC = vertex_positions_mut.size();
    for (int32_t iC = 0; iC < nC; iC++) {
      int32_t iC_prev = (iC + nC - 1) % nC;
      int32_t iC_next = (iC + nC + 1) % nC;
      Float3 pos_prev = vertex_positions_mut[iC_prev];
      Float3 pos = vertex_positions_mut[iC];
      Float3 pos_next = vertex_positions_mut[iC_next];
      Float3 vecA = (pos - pos_prev);
      Float3 vecB = (pos_next - pos);
      Float3 thisN = vecA.cross(vecB);
      plane_normal += thisN;
    }
    plane_normal.normalize();
  }

  // local indices
  std::vector<int32_t> inds(vertex_positions_mut.size());
  std::iota(inds.begin(), inds.end(), 0);

  // NOTE: there are several ways this could be more efficient. It's N^2 thanks
  // to the loop below, but we could use a priority queue, or even just sort an
  // initial list to speed that up. We could also avoid the allocation and
  // tracking of a separate inds buffer.

  while (vertex_positions_mut.size() > 2) {

    // = Find the smallest angle ind
    int32_t smallest_ind = -1;
    double smallest_cosangle = 999999.;
    int32_t nC = vertex_positions_mut.size();
    for (int32_t iC = 0; iC < nC; iC++) {
      int32_t iC_prev = (iC + nC - 1) % nC;
      int32_t iC_next = (iC + nC + 1) % nC;
      Float3 pos_prev = vertex_positions_mut[iC_prev];
      Float3 pos = vertex_positions_mut[iC];
      Float3 pos_next = vertex_positions_mut[iC_next];

      Float3 dirA = (pos - pos_prev).normalized();
      Float3 dirB = (pos_next - pos).normalized();

      double raw_cosangle = dirA.dot(dirB);
      bool convex = dirA.cross(dirB).dot(plane_normal) > 0;
      double cosangle = convex ? raw_cosangle : 2. - raw_cosangle;

      if (smallest_ind == -1 || cosangle < smallest_cosangle) {
        smallest_ind = iC;
        smallest_cosangle = cosangle;
      }
    }
    // = clip the smallest ear

    // local indices
    int32_t iC = smallest_ind;
    int32_t iC_prev = (iC + nC - 1) % nC;
    int32_t iC_next = (iC + nC + 1) % nC;

    // indices into the original triangle
    int32_t i = inds[iC];
    int32_t i_prev = inds[iC_prev];
    int32_t i_next = inds[iC_next];

    // push the new triangle
    tri_inds_out.push_back(Int3{i_prev, i, i_next});

    // remove the clipped vertex from the lists
    inds.erase(inds.begin() + iC);
    vertex_positions_mut.erase(vertex_positions_mut.begin() + iC);
  }
}

Eigen::VectorXi to_eigen(const std::vector<size_t>& vec) {
  Eigen::VectorXi out(vec.size());
  for (size_t i = 0; i < vec.size(); i++) {
    out(i) = vec[i];
  }
  return out;
}

std::tuple<Eigen::VectorXi, Eigen::VectorXi> detect_boundary_loops_helper(gc::surface::ManifoldSurfaceMesh& mesh) {

  // accumulate output here, in a flattened list representation same as
  // used elsewhere in this function
  std::vector<size_t> boundary_loop_inds_flat;
  std::vector<size_t> boundary_loop_degrees = {0};

  // walk the boundary loops to write out their indices
  for (gc::surface::BoundaryLoop bl : mesh.boundaryLoops()) {
    size_t deg = 0;
    for (gc::surface::Vertex v : bl.adjacentVertices()) {
      boundary_loop_inds_flat.push_back(v.getIndex());
      deg++;
    }
    boundary_loop_degrees.push_back(deg);
  }

  return std::tuple<Eigen::VectorXi, Eigen::VectorXi>(to_eigen(boundary_loop_inds_flat),
                                                      to_eigen(boundary_loop_degrees));
}

// convert the nested list format used in geometry-cental
std::vector<std::vector<size_t>> flat_poly_list_to_gc(const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees) {
  size_t nF = F_degrees.rows() - 1;
  std::vector<std::vector<size_t>> nested_faces(nF);
  size_t flat_ind = 0;
  for (size_t iF = 0; iF < nF; iF++) {
    size_t deg = F_degrees(iF + 1);
    for (size_t j = 0; j < deg; j++) {
      nested_faces[iF].push_back(F_flat(flat_ind));
      flat_ind++;
    }
  }
  return nested_faces;
}

std::tuple<Eigen::VectorXi, Eigen::VectorXi> gc_mesh_to_flat_poly(gc::surface::SurfaceMesh& mesh) {

  std::vector<size_t> face_inds_flat;
  std::vector<size_t> face_degrees = {0};

  // walk the faces to write out their indices
  for (gc::surface::Face f : mesh.faces()) {
    size_t deg = 0;
    for (gc::surface::Vertex v : f.adjacentVertices()) {
      face_inds_flat.push_back(v.getIndex());
      deg++;
    }
    face_degrees.push_back(deg);
  }

  return std::tuple<Eigen::VectorXi, Eigen::VectorXi>(to_eigen(face_inds_flat), to_eigen(face_degrees));
}

Eigen::MatrixXi gc_mesh_to_eigen_faces(gc::surface::ManifoldSurfaceMesh& mesh) {
  return mesh.getFaceVertexMatrix<int32_t>();
}

Eigen::MatrixXd gc_mesh_to_eigen_verts(gc::surface::VertexPositionGeometry& geom) {
  return gc::EigenMap<double, 3, Eigen::RowMajor>(geom.vertexPositions);
}

std::vector<gc::Vector3> eigen_to_gc_vector_of_vec3(const Eigen::MatrixXd& vec) {
  std::vector<gc::Vector3> out(vec.rows());
  for (size_t i = 0; i < out.size(); i++) {
    out[i] = gc::Vector3{vec(i, 0), vec(i, 1), vec(i, 2)};
  }
  return out;
}

// ======================================================================
// ==== core functions and exposed bindings
// ======================================================================

NB_MODULE(meshgen_tools_ext, m) {

  m.def(
      "count_self_intersections",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F) {
        Eigen::VectorXi EI;
        Eigen::MatrixXd EV;
        Eigen::MatrixXi IF, EE;

        igl::fast_find_self_intersections(V, F, true, false, IF, EV, EE, EI);

        int64_t n_intersect = IF.rows();

        return n_intersect;
      },
      "V"_a, "F"_a);

  m.def(
      "polygons_flat_list_to_nested_list",
      [](const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees) {
        int32_t nPoly = F_degrees.rows() - 1;
        std::vector<std::vector<int>> nested_faces(nPoly);

        int32_t degSum = 0;
        for (int32_t iF = 0; iF < nPoly; iF++) {
          std::vector<int>& face = nested_faces[iF];
          int32_t D = F_degrees(iF + 1);
          face.resize(D);
          for (int32_t j = 0; j < D; j++) {
            face[j] = F_flat(degSum + j);
          }

          degSum += D;
        }

        return nested_faces;
      },
      "F_flat"_a, "F_degrees"_a);

  m.def(
      "polygons_nested_list_to_flat_list",
      [](const std::vector<std::vector<int>>& nested_faces) {
        int32_t nPoly = nested_faces.size();
        int32_t nCorner = 0;
        for (const std::vector<int>& face : nested_faces) {
          nCorner += face.size();
        }

        Eigen::VectorXi F_flat(nCorner);
        Eigen::VectorXi F_degrees(nPoly + 1);

        int32_t degSum = 0;
        F_degrees(0) = degSum;
        for (int32_t iF = 0; iF < nPoly; iF++) {
          int32_t D = nested_faces[iF].size();
          for (int32_t ind : nested_faces[iF]) {
            F_flat(degSum) = ind;
            degSum++;
          }
          F_degrees(iF + 1) = D;
        }

        return std::tuple<Eigen::VectorXi, Eigen::VectorXi>(F_flat, F_degrees);
      },
      "F_nested"_a);

  // given polygons in a flat list format like
  //   F_flat = [0,1,2,3,4,5,6,7,8,9,10] # flattened list of polygon corner
  //   indices F_degrees = [0, 3, 3, 4]  # degree of each face, padded with an
  //   initial zero
  // this function applies ear-clipping to give a triangulation
  m.def(
      "triangulate_polygons_flat_list",
      [](const Eigen::MatrixXd& V, const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees) {
        int64_t n_polygons = F_degrees.rows() - 1;
        int64_t n_corners = F_flat.rows();
        int64_t n_triangles = n_corners - 2 * n_polygons;

        int64_t iT_out = 0;
        Eigen::MatrixXi F_out = -Eigen::MatrixXi::Ones(n_triangles, 3);

        // walk the faces
        int64_t f_start = 0;
        std::vector<Vector3> vertex_positions; // temp storage
        std::vector<int64_t> corner_inds;      // temp storage
        std::vector<Int3> tri_inds_out;        // temp storate
        for (int64_t iF = 0; iF < n_polygons; iF++) {

          // gather values
          vertex_positions.clear();
          corner_inds.clear();
          tri_inds_out.clear();
          int64_t f_degree = F_degrees[iF + 1];

          // assemble a vector of vertex positions for the polygon corners
          for (int64_t iC = 0; iC < f_degree; iC++) {
            int64_t iV = F_flat[f_start + iC];
            Vector3 v_pos{V(iV, 0), V(iV, 1), V(iV, 2)};
            corner_inds.push_back(iV);
            vertex_positions.push_back(v_pos);
          }

          // run the tesselation subroutine
          std::vector<Vector3> vertex_positions_tmp = vertex_positions; // defensive coding, clipper may mutate this
          triangulate_polygon_ear_clip_mapbox(vertex_positions_tmp, tri_inds_out);
          if (tri_inds_out.size() < (f_degree - 2)) {
            std::cout << "MESHGEN TRIANGULATOR: face " << iF << " fell back on greedy triangulation" << std::endl;
            // In cases with crossing/degenerate polygons, the above algorihtm sometimes
            // outputs fewer than expected triangles or omits regions of the polygon.
            // When this happens, re-run with the simpler greedy algorithm
            tri_inds_out.clear();
            std::vector<Float3> vertex_positions_tmpflt(vertex_positions.size());
            for (int32_t i = 0; i < vertex_positions.size(); i++) {
              for (int32_t j = 0; j < 3; j++) {
                vertex_positions_tmpflt[i][j] = vertex_positions[i][j];
              }
            }
            triangulate_polygon_ear_clip_greedy(vertex_positions_tmpflt, tri_inds_out);
          }

          // Copy the result back (and translate indices)
          for (const Int3& local_face : tri_inds_out) {
            Int3 global_face = {static_cast<int32_t>(corner_inds[local_face[0]]),
                                static_cast<int32_t>(corner_inds[local_face[1]]),
                                static_cast<int32_t>(corner_inds[local_face[2]])};

            F_out.row(iT_out) = global_face;
            iT_out++;
          }

          f_start += f_degree;
        }

        // This _should_ never have any effect, unless the ear clipping fails and outputs fewer triangles. This saves us
        // from outputting garbage in that case.
        if (F_out.rows() > iT_out) {
          F_out = F_out.topRows(iT_out).eval();
        }

        return F_out;
      },
      "V"_a, "F_flat"_a, "F_degrees"_a);

  // Given a set of face indices for a mesh, detect boundary loops and return
  // the ordered lists of vertex indices along the boundary. The boundaries are
  // in-general polygons, are are returned as a tuple of two integer arrays
  // `<vertex_inds_flat, face_degrees>`, in the same polygonal format as is used
  // elsewhere in this class.
  //
  // This version operates on a rectangular face list (pure triangle Fx3 or pure
  // quad mesh Fx4). See below for a polygonal version.
  m.def(
      "detect_boundary_loops",
      [](const Eigen::MatrixXi& F) {
        // construct a geometry-central halfedge mesh
        // if the input mesh is nonmanifold or has other problems, this will
        // throw an exception
        gc::surface::ManifoldSurfaceMesh mesh(F);

        return detect_boundary_loops_helper(mesh);
      },
      "F"_a);

  // Given a set of face indices for a mesh, detect boundary loops and return
  // the ordered lists of vertex indices along the boundary. The boundaries are
  // in-general polygons, are are returned as a tuple of two integer arrays
  // `<vertex_inds_flat, face_degrees>`, in the same polygonal format as is used
  // elsewhere in this class.
  //
  // This version operates on a flattned polygonal face list. See above for a
  // tri/quad version.
  m.def(
      "detect_boundary_loops_poly",
      [](const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees) {
        // construct a geometry-central halfedge mesh
        // if the input mesh is nonmanifold or has other problems, this will
        // throw an exception
        gc::surface::ManifoldSurfaceMesh mesh(flat_poly_list_to_gc(F_flat, F_degrees));

        return detect_boundary_loops_helper(mesh);
      },
      "F_flat"_a, "F_degrees"_a);


  m.def(
      "compute_poly_mesh_stats",
      [](const Eigen::MatrixXd& V, const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees) {
        int32_t nCorners = F_flat.rows();
        int32_t nFaces = F_degrees.rows() - 1;

        // Build an edge lookup structure
        auto edge_key = [](int32_t a, int32_t b) {
          return std::pair<int32_t, int32_t>(std::min(a, b), std::max(a, b));
        };
        std::map<std::pair<int32_t, int32_t>, int32_t> edge_inds;
        auto n_edges = [&]() { return edge_inds.size(); };
        auto edge_ind = [&](int32_t a, int32_t b) { // lazily creates edges
          std::pair<int32_t, int32_t> key = edge_key(a, b);
          if (edge_inds.find(key) == edge_inds.end()) {
            edge_inds[key] = n_edges();
          }
          return edge_inds[key];
        };
        int64_t f_start = 0;
        for (int64_t iF = 0; iF < nFaces; iF++) {
          int64_t f_degree = F_degrees[iF + 1];
          for (int64_t iC = 0; iC < f_degree; iC++) {
            int32_t iCn = (iC + 1) % f_degree;
            int32_t iV = F_flat[f_start + iC];
            int32_t iVn = F_flat[f_start + iCn];
            edge_ind(iV, iVn); // create all edges if we haven't already seen them
          }
          f_start += f_degree;
        }

        Eigen::VectorXd edge_lengths = Eigen::VectorXd::Zero(n_edges());
        Eigen::VectorXi edge_incident_face_counts = Eigen::VectorXi::Zero(n_edges());
        Eigen::VectorXd corner_angles = Eigen::VectorXd::Zero(nCorners);
        Eigen::VectorXi poly_degrees = F_degrees.tail(nFaces);

        f_start = 0;
        for (int64_t iF = 0; iF < nFaces; iF++) {
          int64_t f_degree = F_degrees[iF + 1];
          for (int64_t iC = 0; iC < f_degree; iC++) {
            int32_t iCn = (iC + 1) % f_degree;
            int32_t iCp = (iC - 1 + f_degree) % f_degree;

            int32_t iV = F_flat[f_start + iC];
            int32_t iVn = F_flat[f_start + iCn];
            int32_t iVp = F_flat[f_start + iCp];

            Float3 pV = V.row(iV);
            Float3 pVn = V.row(iVn);
            Float3 pVp = V.row(iVp);

            int32_t iE = edge_ind(iV, iVn);

            edge_lengths(iE) = (pVn - pV).norm();
            edge_incident_face_counts(iE)++;

            Float3 vecA = (pVn - pV);
            Float3 vecB = (pVp - pV);
            vecA.normalize();
            vecB.normalize();
            double dotval = vecA.dot(vecB);
            double cAngle = std::acos(dotval);
            corner_angles(f_start + iC) = cAngle; // WARNING: if anyle is > 180 deg, this measure exterior angle
          }
          f_start += f_degree;
        }

        return std::make_tuple(edge_lengths, edge_incident_face_counts, corner_angles, poly_degrees);
      },
      "V"_a, "F_flat"_a, "F_degrees"_a);

  // Bind the options struct
  nb::class_<MeshAugmentorOptions>(m, "MeshAugmentorOptions")
      .def(nb::init<>())
      .def_rw("face_subdiv_rate", &MeshAugmentorOptions::face_subdiv_rate)
      .def_rw("edge_split_rate", &MeshAugmentorOptions::edge_split_rate)
      .def_rw("planar_edge_flip_rate", &MeshAugmentorOptions::planar_edge_flip_rate)
      .def_rw("planar_threshold", &MeshAugmentorOptions::planar_threshold)
      .def_rw("rel_eps", &MeshAugmentorOptions::rel_eps);


  m.def(
      "augment_mesh",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F, const MeshAugmentorOptions& opts) {
        // construct a geometry-central halfedge mesh
        // if the input mesh is nonmanifold or has other problems, this will
        // throw an exception
        std::unique_ptr<gc::surface::ManifoldSurfaceMesh> mesh;
        std::unique_ptr<gc::surface::VertexPositionGeometry> geom;
        std::tie(mesh, geom) = gc::surface::makeManifoldSurfaceMeshAndGeometry(V, F);
        if (!mesh->isTriangular()) {
          throw std::runtime_error("input is not a manifold triangular mesh");
        }

        MeshAugmentor augmentor(*mesh, *geom, opts);
        augmentor.run();

        return std::tuple<Eigen::MatrixXd, Eigen::MatrixXi>(gc_mesh_to_eigen_verts(*geom),
                                                            gc_mesh_to_eigen_faces(*mesh));
      },
      "V"_a, "F"_a, "Opts"_a);

  m.def(
      "augment_mesh_poly",
      [](const Eigen::MatrixXd& V, const Eigen::VectorXi& F_flat, const Eigen::VectorXi& F_degrees,
         const MeshAugmentorOptions& opts) {
        // construct a geometry-central halfedge mesh
        // if the input mesh is nonmanifold or has other problems, this will
        // throw an exception
        std::unique_ptr<gc::surface::ManifoldSurfaceMesh> mesh;
        std::unique_ptr<gc::surface::VertexPositionGeometry> geom;
        std::tie(mesh, geom) = gc::surface::makeManifoldSurfaceMeshAndGeometry(flat_poly_list_to_gc(F_flat, F_degrees),
                                                                               eigen_to_gc_vector_of_vec3(V));

        MeshAugmentor augmentor(*mesh, *geom, opts);
        augmentor.run();

        std::tuple<Eigen::VectorXi, Eigen::VectorXi> tup = gc_mesh_to_flat_poly(*mesh);

        return std::tuple<Eigen::MatrixXd, Eigen::VectorXi, Eigen::VectorXi>(gc_mesh_to_eigen_verts(*geom),
                                                                             std::get<0>(tup), std::get<1>(tup));
      },
      "V"_a, "F_flat"_a, "F_deg"_a, "Opts"_a);

  m.def(
      "load_mesh_poly_flat",
      [](std::string filename) {
        // construct a geometry-central halfedge mesh
        // if the input mesh is nonmanifold or has other problems, this will
        // throw an exception
        std::unique_ptr<gc::surface::SurfaceMesh> mesh;
        std::unique_ptr<gc::surface::VertexPositionGeometry> geom;
        std::tie(mesh, geom) = gc::surface::readSurfaceMesh(filename);

        std::tuple<Eigen::VectorXi, Eigen::VectorXi> tup = gc_mesh_to_flat_poly(*mesh);

        return std::tuple<Eigen::MatrixXd, Eigen::VectorXi, Eigen::VectorXi>(gc_mesh_to_eigen_verts(*geom),
                                                                             std::get<0>(tup), std::get<1>(tup));
      },
      "filename"_a);


  m.def(
      "check_is_trivial_quad_split",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F) {
        if (F.rows() % 2 != 0) return false;

        for (int32_t iF = 0; iF < F.rows(); iF += 2) {

          int32_t iFn = iF + 1;

          bool found_match = false;
          for (int32_t j = 0; j < 3; j++) {
            for (int32_t k = 0; k < 3; k++) {
              if ((F(iF, j) == F(iFn, (k + 1) % 3)) && (F(iF, (j + 1) % 3) == F(iFn, k))) {
                // the triangles contain a shared edge with compatible orientation
                found_match = true;
              }
            }
          }

          if (!found_match) return false;
        }

        return true;
      },
      "V"_a, "F"_a);


  m.def(
      "trivial_quad_union",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F) {
        Eigen::MatrixXi F_new(F.rows() / 2, 4);

        for (int32_t iF = 0; iF < F.rows(); iF += 2) {

          int32_t iFn = iF + 1;
          int32_t iFnew = iF / 2;

          bool found_match = false;
          for (int32_t j = 0; j < 3; j++) {
            for (int32_t k = 0; k < 3; k++) {
              if ((F(iF, j) == F(iFn, (k + 1) % 3)) && (F(iF, (j + 1) % 3) == F(iFn, k))) {
                // the triangles contain a shared edge with compatible orientation
                found_match = true;

                F_new(iFnew, 0) = F(iF, (j + 2) % 3);
                F_new(iFnew, 1) = F(iF, (j + 0) % 3);
                F_new(iFnew, 2) = F(iFn, (k + 2) % 3);
                F_new(iFnew, 3) = F(iF, (j + 1) % 3);
              }

              if (found_match) break;
            }
            if (found_match) break;
          }
        }

        return F_new;
      },
      "V"_a, "F"_a);

  m.def(
      "deduplicate_and_reorient_triangle_mesh",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F) {
        if (F.cols() != 3) throw std::runtime_error("only triangle meshes are supported");

        // Convert types
        std::vector<std::vector<size_t>> polygons(F.rows(), std::vector<size_t>(3));
        for (int32_t iF = 0; iF < F.rows(); iF++) {
          for (int32_t j = 0; j < 3; j++) {
            polygons[iF][j] = F(iF, j);
          }
        }
        std::vector<Vector3> vertexCoordinates(V.rows());
        for (int32_t iV = 0; iV < V.rows(); iV++) {
          for (int32_t j = 0; j < 3; j++) {
            vertexCoordinates[iV][j] = V(iV, j);
          }
        }

        // Repair the mesh
        gc::SimplePolygonMesh polygonMesh(polygons, vertexCoordinates);
        polygonMesh.stripUnusedVertices();
        polygonMesh.stripInvalidFaces();
        polygonMesh.stripDuplicateFaces();
        polygonMesh.consistentlyOrientFaces(true);

        // Convert types back
        Eigen::MatrixXd V_new(polygonMesh.vertexCoordinates.size(), 3);
        Eigen::MatrixXi F_new(polygonMesh.polygons.size(), 3);
        for (int32_t iF = 0; iF < F_new.rows(); iF++) {
          for (int32_t j = 0; j < 3; j++) {
            F_new(iF, j) = polygonMesh.polygons[iF][j];
          }
        }
        for (int32_t iV = 0; iV < V_new.rows(); iV++) {
          for (int32_t j = 0; j < 3; j++) {
            V_new(iV, j) = polygonMesh.vertexCoordinates[iV][j];
          }
        }


        return std::tuple<Eigen::MatrixXd, Eigen::MatrixXi>(V_new, F_new);
      },
      "V"_a, "F"_a);
  
  m.def(
      "heat_method_distance",
      [](const Eigen::MatrixXd& V, const Eigen::MatrixXi& F, int32_t sourceVertInd) {

        std::unique_ptr<gc::surface::SurfaceMesh> mesh;
        std::unique_ptr<gc::surface::VertexPositionGeometry> geom;

        std::tie(mesh, geom) = gc::surface::makeSurfaceMeshAndGeometry(V, F);

        gc::Vertex sourceVert = mesh->vertex(sourceVertInd);

        gc::VertexData<double> distToSource = gc::heatMethodDistance(*geom, sourceVert);

        Eigen::VectorXd distsOut = distToSource.toVector();

        return distsOut;
      },
      "V"_a, "F"_a, "sourceVert"_a);

}
