import os
import sys
import os.path as path
import argparse

import numpy as np
import polyscope as ps
import igl

import meshgen_tools



def main():

    parser = argparse.ArgumentParser()

    # Build arguments
    parser.add_argument('mesh', type=str, help='path to a mesh')
    parser.add_argument('--do_viz', action='store_true', help='do viz')

    # Parse arguments
    args = parser.parse_args()

    if args.do_viz:
        ps.init()

    # Load a mesh argument
    V, F = igl.read_triangle_mesh(args.mesh)
    print(f"loaded mesh with {V.shape[0]} verts and {F.shape[0]} faces")
    if args.do_viz:
        ps.register_surface_mesh("input mesh", V, F)
        ps.show()
    
    # Test dedup and reorient
    print("\n## Dedup and Reorient Testing")
    V_new, F_new = meshgen_tools.deduplicate_and_reorient_triangle_mesh(V,F)
    if args.do_viz:
        ps.register_surface_mesh("dedup mesh", V_new, F_new, back_face_policy="custom")
        ps.show()

    # Test the intersections function
    print("\n## Intersection Testing")
    n_intersect = meshgen_tools.count_self_intersections(V,F)
    print(f"found {n_intersect} self-intersections")
    

    # Dummy data
    # poly_V = np.random.rand(10,3)
    # poly_faces_flat = np.array([0,1,2,3,4,5,6,7,8,9])
    # poly_faces_size = np.array([0, 3, 3, 4])

    # Load mesh arg
    poly_V, poly_faces_flat, poly_faces_size = meshgen_tools.load_mesh_poly_flat(args.mesh)


    # Test the ear-clipping triangulation
    # NOTE: this currently ignores the loaded mesh and uses the inline data
    print("\n## Ear Clipping (polygonal)")
    poly_F_flat = meshgen_tools.triangulate_polygons_flat_list(poly_V, poly_faces_flat, poly_faces_size)
    print(poly_F_flat)
    


    # Test stats

    edge_lengths, edge_incident_face_counts, corner_angles, poly_degrees = meshgen_tools.compute_poly_mesh_stats(poly_V, poly_faces_flat, poly_faces_size)
    print("Edge lengths:")
    print(edge_lengths)
    print("Edge incident face counts:")
    print(edge_incident_face_counts)
    print("Corner angles:")
    print(corner_angles)
    print("Poly degrees:")
    print(poly_degrees)
    
    # Smoke test for poly augmentor
    print("\n## Augmentor smoke test (polygonal)")
    opts = meshgen_tools.MeshAugmentorOptions()
    newV, newF_flat, newF_deg = meshgen_tools.augment_mesh_poly(poly_V, poly_faces_flat, poly_faces_size, opts)
    print(newV)
    print(newF_flat)
    print(newF_deg)
        
    # Test the boundary loop detection
    print("\n## Detecting Boundary Loops (pure tri/quad)")
    boundary_face_inds, boundary_face_deg = meshgen_tools.detect_boundary_loops(F)
    print(boundary_face_inds)
    print(boundary_face_deg)
    
    print("\n## Detecting Boundary Loops (polygonal)")
    boundary_face_inds, boundary_face_deg = meshgen_tools.detect_boundary_loops_poly(poly_faces_flat, poly_faces_size)
    print(boundary_face_inds)
    print(boundary_face_deg)


if __name__ == '__main__':
    main()
